@@ -13,10 +13,16 @@ from astropy.coordinates import (
 )
 from astropy.tests.helper import assert_quantity_allclose
 
+from sunpy import sun
+from sunpy.coordinates.frames import (
+    Heliocentric,
+    HeliographicCarrington,
+    HeliographicStonyhurst,
+    Helioprojective,
+)
+from sunpy.coordinates.sun import angular_radius
 from sunpy.time import parse_time
 from sunpy.util.exceptions import SunpyUserWarning
-from ... import sun
-from ..frames import Heliocentric, HeliographicCarrington, HeliographicStonyhurst, Helioprojective
 
 RSUN_METERS = sun.constants.get('radius').si.to(u.m)
 DSUN_METERS = sun.constants.get('mean distance').si.to(u.m)
@@ -222,7 +228,7 @@ def test_hpc_low_precision_float_warning():
                           u.Quantity(0, u.arcsec, dtype=np.float16),
                           observer=HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU))
 
-    with pytest.warns(SunpyUserWarning, match="Tx is float32, and Ty is float16"):
+    with pytest.raises(SunpyUserWarning, match="Tx is float32, and Ty is float16"):
         hpc.make_3d()
 
 
@@ -428,3 +434,20 @@ def test_hgc_incomplete_observer():
     with pytest.raises(ValueError, match=r'Full 3D coordinate \(including radius\) must be specified'):
         SkyCoord(0*u.deg, 0*u.deg, frame="heliographic_carrington",
                  observer='self', obstime="2011-01-01T00:00:00")
+
+
+def test_angular_radius():
+    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime="2010/01/01T00:00:00", observer="earth")
+    assert_quantity_allclose(coord.angular_radius, angular_radius(coord.obstime))
+
+
+def test_angular_radius_no_observer():
+    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime="2010/01/01T00:00:00", observer=None)
+    with pytest.raises(ValueError, match=r"The observer must be defined, not `None`"):
+        coord.angular_radius
+
+
+def test_angular_radius_no_obstime():
+    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime=None, observer="earth")
+    with pytest.raises(ValueError, match=r"The observer must be fully defined by specifying `obstime`."):
+        coord.angular_radius
