@@ -3,33 +3,27 @@ This module provides a wrapper around the VSO API.
 """
 
 import os
-import re
 import cgi
+import copy
 import json
 import socket
 import inspect
 import datetime
-import warnings
 import itertools
+from pathlib import Path
 from functools import partial
-from collections import defaultdict
 from urllib.error import URLError, HTTPError
 from urllib.parse import urlencode
 from urllib.request import Request, urlopen
 
 import zeep
-from zeep.helpers import serialize_object
-
-import astropy.units as u
-from astropy.table import QTable as Table
 
 from sunpy import config, log
 from sunpy.net.attr import and_
-from sunpy.net.base_client import BaseClient, BaseQueryResponse
+from sunpy.net.base_client import BaseClient, QueryResponseRow
 from sunpy.net.vso import attrs
 from sunpy.net.vso.attrs import _walker as walker
-from sunpy.time import TimeRange, parse_time
-from sunpy.util.exceptions import SunpyUserWarning
+from sunpy.util.exceptions import warn_user
 from sunpy.util.net import slugify
 from sunpy.util.parfive_helpers import Downloader, Results
 from .. import _attrs as core_attrs
@@ -42,84 +36,28 @@ from .exceptions import (
     UnknownStatus,
     UnknownVersion,
 )
+from .legacy_response import QueryResponse
+from .table_response import VSOQueryResponseTable
 from .zeep_plugins import SunPyLoggingZeepPlugin
 
-TIME_FORMAT = config.get("general", "time_format")
-
 DEFAULT_URL_PORT = [{'url': 'http://docs.virtualsolar.org/WSDL/VSOi_rpc_literal.wsdl',
                      'port': 'nsoVSOi'},
                     {'url': 'https://sdac.virtualsolar.org/API/VSOi_rpc_literal.wsdl',
                      'port': 'sdacVSOi'}]
 
-RANGE = re.compile(r'(\d+)(\s*-\s*(\d+))?(\s*([a-zA-Z]+))?')
-
 
 class _Str(str):
-
     """ Subclass of string that contains a meta attribute for the
     record_item associated with the file. """
+    meta = None
 
 
 # ----------------------------------------
-
-def _parse_waverange(string):
-    min_, max_, unit = RANGE.match(string).groups()[::2]
-    return {
-        'wave_wavemin': min_,
-        'wave_wavemax': min_ if max_ is None else max_,
-        'wave_waveunit': 'Angstrom' if unit is None else unit,
-    }
-
-
-def _parse_date(string):
-    start, end = string.split(' - ')
-    return {'time_start': start.strip(), 'time_end': end.strip()}
-
-
-def iter_sort_response(response):
-    """
-    Sorts the VSO queryresults by their start time.
-
-    Parameters
-    ----------
-    response : `zeep.objects.QueryResponse`
-        A SOAP Object of a VSO queryresult
-
-    Returns
-    -------
-    `list`
-        Sorted record items w.r.t. their start time.
-    """
-    has_time_recs = list()
-    has_notime_recs = list()
-    for prov_item in response.provideritem:
-        if not hasattr(prov_item, 'record') or not prov_item.record:
-            continue
-        if not hasattr(prov_item.record, 'recorditem') or not prov_item.record.recorditem:
-            continue
-        rec_item = prov_item.record.recorditem
-        for rec in rec_item:
-            if hasattr(rec, 'time') and hasattr(rec.time, 'start') and rec.time.start is not None:
-                has_time_recs.append(rec)
-            else:
-                has_notime_recs.append(rec)
-    has_time_recs = sorted(has_time_recs, key=lambda x: x.time.start)
-    all_recs = has_time_recs + has_notime_recs
-    return all_recs
-
-
-def iter_errors(response):
-    for prov_item in response.provideritem:
-        if not hasattr(prov_item, 'record') or not prov_item.record:
-            yield prov_item
-
-
 def check_connection(url):
     try:
         return urlopen(url).getcode() == 200
     except (socket.error, socket.timeout, HTTPError, URLError) as e:
-        warnings.warn(f"Connection to {url} failed with error {e}. Retrying with different url and port.",
-                      SunpyUserWarning)
+        warn_user(f"Connection to {url} failed with error {e}. Retrying with different url and port.")
         return None
 
 
@@ -140,16 +78,13 @@ def build_client(url=None, port_name=None, **kwargs):
     ----------
     url : `str`
         The URL to connect to.
-
     port_name : `str`
         The "port" to use.
-
-    kwargs : `dict`
+    **kwargs : `dict`
         All extra keyword arguments are passed to `zeep.Client`.
 
     Returns
     -------
-
     `zeep.Client`
     """
     if url is None and port_name is None:
@@ -172,149 +107,9 @@ def build_client(url=None, port_name=None, **kwargs):
     return client
 
 
-class QueryResponse(BaseQueryResponse):
-    """
-    A container for VSO Records returned from VSO Searches.
-    """
-
-    def __init__(self, lst, queryresult=None):
-        super().__init__()
-        self._data = lst
-        self.queryresult = queryresult
-        self.errors = []
-        self._client = VSOClient()
-
-    def __getitem__(self, item):
-        # Always index so a list comes back
-        if isinstance(item, int):
-            item = slice(item, item+1)
-        return type(self)(self._data[item], queryresult=self.queryresult)
-
-    def __len__(self):
-        return len(self._data)
-
-    def __iter__(self):
-        for block in self._data:
-            yield block
-
-    @property
-    def blocks(self):
-        return self._data
-
-    @property
-    def client(self):
-        return self._client
-
-    @client.setter
-    def client(self, client):
-        self._client = client
-
-    def search(self, *query):
-        """ Furtherly reduce the query response by matching it against
-        another query, e.g. response.search(attrs.Instrument('aia')). """
-        query = and_(*query)
-        return QueryResponse(
-            attrs._filter_results(query, self), self.queryresult
-        )
-
-    @classmethod
-    def create(cls, queryresult):
-        res = list(iter_sort_response(queryresult))
-        return cls(res, queryresult)
-
-    def total_size(self):
-        """ Total size of data in KB. May be less than the actual
-        size because of inaccurate data providers. """
-        # Warn about -1 values?
-        return sum(record.size for record in self if record.size > 0)
-
-    def time_range(self):
-        """ Return total time-range all records span across. """
-        return TimeRange(min(record.time.start for record in self if record.time.start is not None),
-                         max(record.time.end for record in self if record.time.end is not None))
-
-    def build_table(self):
-        """
-        Create a human readable table.
-
-        Returns
-        -------
-        table : `astropy.table.QTable`
-        """
-        keywords = ['Start Time', 'End Time', 'Source', 'Instrument', 'Type', 'Wavelength']
-        record_items = {}
-        for key in keywords:
-            record_items[key] = []
-
-        def validate_time(time):
-            # Handle if the time is None when coming back from VSO
-            if time is None:
-                return ['None']
-            if record.time.start is not None:
-                return [parse_time(time).strftime(TIME_FORMAT)]
-            else:
-                return ['N/A']
-
-        for record in self:
-            record_items['Start Time'].append(validate_time(record.time.start))
-            record_items['End Time'].append(validate_time(record.time.end))
-            record_items['Source'].append(str(record.source))
-            record_items['Instrument'].append(str(record.instrument))
-            record_items['Type'].append(str(record.extent.type)
-                                        if record.extent.type is not None else ['N/A'])
-            # If we have a start and end Wavelength, make a quantity
-            if hasattr(record, 'wave') and record.wave.wavemin and record.wave.wavemax:
-                unit = record.wave.waveunit
-                # Convert this so astropy units parses it correctly
-                if unit == "kev":
-                    unit = "keV"
-                record_items['Wavelength'].append(u.Quantity([float(record.wave.wavemin),
-                                                              float(record.wave.wavemax)],
-                                                             unit=unit))
-            # If not save None
-            else:
-                record_items['Wavelength'].append(None)
-        # If we have no wavelengths for the whole list, drop the col
-        if all([a is None for a in record_items['Wavelength']]):
-            record_items.pop('Wavelength')
-            keywords.remove('Wavelength')
-        else:
-            # Make whole column a quantity
-            try:
-                with u.set_enabled_equivalencies(u.spectral()):
-                    record_items['Wavelength'] = u.Quantity(record_items['Wavelength'])
-            # If we have mixed units or some Nones just represent as strings
-            except (u.UnitConversionError, TypeError):
-                record_items['Wavelength'] = [str(a) for a in record_items['Wavelength']]
-
-        return Table(record_items)[keywords]
-
-    def add_error(self, exception):
-        self.errors.append(exception)
-
-    def response_block_properties(self):
-        """
-        Returns a set of class attributes on all the response blocks.
-
-        Returns
-        -------
-        s : `set`
-            List of strings, containing attribute names in the response blocks.
-        """
-        s = {a if not a.startswith('_') else None for a in dir(self[0])}
-        for resp in self[1:]:
-            if len(s) == 0:
-                break
-            s = s.intersection({a if not a.startswith('_') else None for a in dir(resp)})
-
-        if None in s:
-            s.remove(None)
-        return s
-
-
 class VSOClient(BaseClient):
     """
-    Allows queries and downloads from the Virtual Solar Observatory (VSO).
+    Provides access to query and download from Virtual Solar Observatory (VSO).
 
     Parameters
     ----------
@@ -339,6 +134,18 @@ class VSOClient(BaseClient):
                 raise ConnectionError("Cannot find an online VSO mirror.")
         self.api = api
 
+    def __deepcopy__(self, memo):
+        """
+        Copy the client but don't copy the API object.
+        """
+        memo[id(self.api)] = self.api
+        deepcopy_method = self.__deepcopy__
+        self.__deepcopy__ = None
+        cp = copy.deepcopy(self, memo)
+        self.__deepcopy__ = deepcopy_method
+        cp.__deepcopy__ = deepcopy_method
+        return cp
+
     def make(self, atype, **kwargs):
         """
         Create a new SOAP object.
@@ -346,11 +153,18 @@ class VSOClient(BaseClient):
         obj = self.api.get_type(f"VSO:{atype}")
         return obj(**kwargs)
 
-    def search(self, *query):
-        """ Query data from the VSO with the new API. Takes a variable number
+    def search(self, *query, response_format=None):
+        """
+        Query data from the VSO with the new API. Takes a variable number
         of attributes as parameter, which are chained together using AND.
 
-        The new query language allows complex queries to be easily formed.
+        Parameters
+        ----------
+        response_format : {``"legacy"``, ``"table"``}, optional
+            The response format from the search, this can be either
+            ``"legacy"`` to return a list-like object of the zeep responses, or
+            ``"table"`` to return the responses in a subclass of
+            `~astropy.table.QTable`.
 
         Examples
         --------
@@ -362,27 +176,31 @@ class VSOClient(BaseClient):
         >>> client = vso.VSOClient()  # doctest: +REMOTE_DATA
         >>> client.search(
         ...    a.Time(datetime(2010, 1, 1), datetime(2010, 1, 1, 1)),
-        ...    a.Instrument.eit | a.Instrument.aia)   # doctest:  +REMOTE_DATA
-        <sunpy.net.vso.vso.QueryResponse object at ...>
-           Start Time [1]       End Time [1]    Source ...   Type   Wavelength [2]
-                                                       ...             Angstrom
-        ------------------- ------------------- ------ ... -------- --------------
-        2010-01-01 00:00:08 2010-01-01 00:00:20   SOHO ... FULLDISK 195.0 .. 195.0
-        2010-01-01 00:12:08 2010-01-01 00:12:20   SOHO ... FULLDISK 195.0 .. 195.0
-        2010-01-01 00:24:10 2010-01-01 00:24:22   SOHO ... FULLDISK 195.0 .. 195.0
-        2010-01-01 00:36:08 2010-01-01 00:36:20   SOHO ... FULLDISK 195.0 .. 195.0
-        2010-01-01 00:48:09 2010-01-01 00:48:21   SOHO ... FULLDISK 195.0 .. 195.0
+        ...    a.Instrument.eit | a.Instrument.aia,
+        ...    response_format="table")   # doctest:  +REMOTE_DATA
+        <sunpy.net.vso.table_response.VSOQueryResponseTable object at ...>
+            Start Time               End Time        Source ... Extent Type   Size
+                                                            ...              Mibyte
+        ----------------------- ----------------------- ------ ... ----------- -------
+        2010-01-01 00:00:08.000 2010-01-01 00:00:20.000   SOHO ...    FULLDISK 2.01074
+        2010-01-01 00:12:08.000 2010-01-01 00:12:20.000   SOHO ...    FULLDISK 2.01074
+        2010-01-01 00:24:10.000 2010-01-01 00:24:22.000   SOHO ...    FULLDISK 2.01074
+        2010-01-01 00:36:08.000 2010-01-01 00:36:20.000   SOHO ...    FULLDISK 2.01074
+        2010-01-01 00:48:09.000 2010-01-01 00:48:21.000   SOHO ...    FULLDISK 2.01074
 
         Returns
         -------
-        out : :py:class:`QueryResult` (enhanced list)
+        out : `~sunpy.net.vso.table_response.VSOQueryResponseTable`
             Matched items. Return value is of same type as the one of
-            :py:meth:`VSOClient.search`.
+            :meth:`VSOClient.search`.
         """
+        if response_format is None:
+            response_format = "table"
         query = and_(*query)
         QueryRequest = self.api.get_type('VSO:QueryRequest')
         VSOQueryResponse = self.api.get_type('VSO:QueryResponse')
         responses = []
+        exceptions = []
         for block in walker.create(query, self.api):
             try:
                 query_response = self.api.service.Query(
@@ -390,15 +208,23 @@ class VSOClient(BaseClient):
                 )
                 for resp in query_response:
                     if resp["error"]:
-                        warnings.warn(resp["error"], SunpyUserWarning)
+                        warn_user(resp["error"])
                 responses.append(
                     VSOQueryResponse(query_response)
                 )
             except Exception as ex:
-                response = QueryResponse.create(self.merge(responses))
-                response.add_error(ex)
+                exceptions.append(ex)
 
-        return QueryResponse.create(self.merge(responses))
+        responses = self.merge(responses)
+        if response_format == "legacy":
+            response = QueryResponse.create(responses)
+        else:
+            response = VSOQueryResponseTable.from_zeep_response(responses, client=self)
+
+        for ex in exceptions:
+            response.add_error(ex)
+
+        return response
 
     def merge(self, queryresponses):
         """ Merge responses into one. """
@@ -434,36 +260,30 @@ class VSOClient(BaseClient):
                          provideritem=list(providers.values()))
 
     @staticmethod
-    def mk_filename(pattern, queryresponse, resp, url):
+    def mk_filename(pattern, queryresponserow, resp, url):
         """
         Generate the best possible (or least-worse) filename for a VSO download.
 
         * Use the ``content-disposition`` header.
-        * Use `fileid` to generate a file name if content-disposition fails
+        * Use ``fileid`` to generate a file name if content-disposition fails
         * If everything else fails use the last segment of the URL and hope.
         """
         name = None
         if resp:
             cdheader = resp.headers.get("Content-Disposition", None)
             if cdheader:
-                value, params = cgi.parse_header(cdheader)
+                _, params = cgi.parse_header(cdheader)
                 name = params.get('filename', "")
                 # Work around https://github.com/sunpy/sunpy/issues/3372
                 if name.count('"') >= 2:
                     name = name.split('"')[1]
 
         if name is None:
-            # Advice from the VSO is to fallback to providerid + fileid
+            # Advice from the VSO is to fallback to providerid + fileid for a filename
             # As it's possible multiple providers give the same fileid.
             # However, I haven't implemented this yet as it would be a breaking
             # change to the filenames we expect.
-
-            # I don't know if we still need this bytes check in Python 3 only
-            # land, but I don't dare remove it.
-            if isinstance(queryresponse.fileid, bytes):
-                fileid = queryresponse.fileid.decode("ascii", "ignore")
-            else:
-                fileid = queryresponse.fileid
+            fileid = queryresponserow['fileid']
 
             # Some providers make fileid a path
             # Some also don't specify a file extension, but not a lot we can do
@@ -481,7 +301,8 @@ class VSOClient(BaseClient):
         if not name:
             name = f"vso_file_{datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')}"
 
-        fname = pattern.format(file=name, **serialize_object(queryresponse))
+        fname = pattern.format(file=name,
+                               **queryresponserow.response_block_map)
 
         return fname
 
@@ -492,25 +313,22 @@ class VSOClient(BaseClient):
 
         Parameters
         ----------
-        query_response : sunpy.net.vso.QueryResponse
+        query_response : sunpy.net.vso.VSOQueryResponseTable
             QueryResponse containing the items to be downloaded.
-
         path : str
             Specify where the data is to be downloaded. Can refer to arbitrary
             fields of the QueryResponseItem (instrument, source, time, ...) via
             string formatting, moreover the file-name of the file downloaded can
             be referred to as file, e.g.
             "{source}/{instrument}/{time.start}/{file}".
-
-        methods : {list of str}
+        methods : `list` of `str`
             Download methods, defaults to URL-FILE_Rice then URL-FILE.
             Methods are a concatenation of one PREFIX followed by any number of
-            SUFFIXES i.e. `PREFIX-SUFFIX_SUFFIX2_SUFFIX3`.
+            SUFFIXES i.e. ``PREFIX-SUFFIX_SUFFIX2_SUFFIX3``.
             The full list of
             `PREFIXES <https://sdac.virtualsolar.org/cgi/show_details?keyword=METHOD_PREFIX>`_
             and `SUFFIXES <https://sdac.virtualsolar.org/cgi/show_details?keyword=METHOD_SUFFIX>`_
             are listed on the VSO site.
-
         site : str
             There are a number of caching mirrors for SDO and other
             instruments, some available ones are listed below.
@@ -530,20 +348,17 @@ class VSOClient(BaseClient):
         progress : `bool`, optional
             If `True` show a progress bar showing how many of the total files
             have been downloaded. If `False`, no progress bars will be shown at all.
-
         overwrite : `bool` or `str`, optional
             Determine how to handle downloading if a file already exists with the
             same name. If `False` the file download will be skipped and the path
             returned to the existing file, if `True` the file will be downloaded
-            and the existing file will be overwritten, if `'unique'` the filename
+            and the existing file will be overwritten, if ``'unique'`` the filename
             will be modified to be unique.
-
         downloader : `parfive.Downloader`, optional
             The download manager to use.
-
         wait : `bool`, optional
-           If `False` ``downloader.download()`` will not be called. Only has
-           any effect if `downloader` is not `None`.
+            If `False` ``downloader.download()`` will not be called. Only has
+            any effect if ``downloader`` is not `None`.
 
         Returns
         -------
@@ -555,20 +370,28 @@ class VSOClient(BaseClient):
         >>> files = fetch(qr) # doctest:+SKIP
         """
         if path is None:
-            path = os.path.join(config.get('downloads', 'download_dir'),
-                                '{file}')
-        elif isinstance(path, str) and '{file}' not in path:
-            path = os.path.join(path, '{file}')
-        path = os.path.expanduser(path)
+            path = Path(config.get('downloads', 'download_dir')) / '{file}'
+        elif isinstance(path, (str, os.PathLike)) and '{file}' not in str(path):
+            path = Path(path) / '{file}'
+        else:
+            path = Path(path)
+        path = path.expanduser()
 
         dl_set = True
         if not downloader:
             dl_set = False
-            downloader = Downloader(progress=progress)
+            downloader = Downloader(progress=progress, overwrite=overwrite)
 
-        fileids = VSOClient.by_fileid(query_response)
-        if not fileids:
+        if isinstance(query_response, (QueryResponse, list)):
+            query_response = VSOQueryResponseTable.from_zeep_response(query_response,
+                                                                      client=self,
+                                                                      _sort=False)
+        if isinstance(query_response, QueryResponseRow):
+            query_response = query_response.as_table()
+
+        if not len(query_response):
             return downloader.download() if wait else Results()
+
         # Adding the site parameter to the info
         info = {}
         if site is not None:
@@ -579,7 +402,11 @@ class VSOClient(BaseClient):
         data_request = self.make_getdatarequest(query_response, methods, info)
         data_response = VSOGetDataResponse(self.api.service.GetData(data_request))
 
-        err_results = self.download_all(data_response, methods, downloader, path, fileids)
+        err_results = self.download_all(data_response,
+                                        methods,
+                                        downloader,
+                                        str(path),
+                                        self.by_fileid(query_response))
 
         if dl_set and not wait:
             return err_results
@@ -589,37 +416,19 @@ class VSOClient(BaseClient):
         results._errors += err_results.errors
         return results
 
-    @staticmethod
-    def link(query_response, maps):
-        """ Return list of paths with records associated with them in
-        the meta attribute. """
-        if not maps:
-            return []
-        ret = []
-
-        for record_item in query_response:
-            try:
-                item = _Str(maps[record_item.fileid]['path'])
-            except KeyError:
-                continue
-            item.meta = record_item
-            ret.append(item)
-        return ret
-
     def make_getdatarequest(self, response, methods=None, info=None):
         """ Make datarequest with methods from response. """
         if methods is None:
             methods = self.method_order + ['URL']
 
         return self.create_getdatarequest(
-            {k: [x.fileid for x in v]
-             for k, v in self.by_provider(response).items()},
+            {g[0]['Provider']: list(g['fileid']) for g in response.group_by('Provider').groups},
             methods, info
         )
 
     def create_getdatarequest(self, maps, methods, info=None):
         """ Create datarequest from maps mapping data provider to
-        fileids and methods, """
+        fileids and methods,"""
         if info is None:
             info = {}
 
@@ -742,18 +551,6 @@ class VSOClient(BaseClient):
 
         raise NoData
 
-    @staticmethod
-    def by_provider(response):
-        """
-        Returns a dictionary of provider
-        corresponding to records in the response.
-        """
-
-        map_ = defaultdict(list)
-        for record in response:
-            map_[record.provider].append(record)
-        return map_
-
     @staticmethod
     def by_fileid(response):
         """
@@ -761,7 +558,7 @@ class VSOClient(BaseClient):
         corresponding to records in the response.
         """
         return {
-            record.fileid: record for record in response
+            record['fileid']: record for record in response
         }
 
     def multiple_choices(self, choices, response):
@@ -864,4 +661,8 @@ class VSOClient(BaseClient):
                         attrs[key].append((str(item[key]), str(item[key+"_long"])))
 
         with open(os.path.join(here, 'data', 'attrs.json'), 'w') as attrs_file:
-            json.dump(attrs, attrs_file)
+            json.dump(dict(sorted(attrs.items())), attrs_file, indent=2)
+
+    @property
+    def info_url(self):
+        return 'http://vso.stanford.edu/cgi-bin/search'
